@@ -1,5 +1,17 @@
 /* jshint esversion: 6 */
 
+/*-------*
+ * Setup *
+ *-------*/
+
+function doSetup(f) {
+    if (document.readyState == "complete") {
+        f();
+    } else {
+        window.addEventListener("load", f);
+    }
+}
+
 /*--------------------*
  * Code line focusing *
  *--------------------*/
@@ -154,7 +166,7 @@ function focusDestinationLine() {
 
 
 // Add click handlers to each code line and each lineref link
-window.addEventListener("load", function () {
+doSetup(function() {
     let codelines = document.querySelectorAll("span.codeline");
     for (let line of codelines) {
         line.addEventListener("click", focusThisCodeLine);
@@ -214,7 +226,7 @@ function toggleTab() {
 
 
 // Add click handlers to every tab
-window.addEventListener("load", function () {
+doSetup(function() {
     let tab_buttons = document.querySelectorAll("div.tabs li.tab");
     for (let button of tab_buttons) {
         button.addEventListener("click", toggleTab);
@@ -241,12 +253,12 @@ function hide_accomplished_goals(hide) {
         if (hide) {
             row.style.display = "none";
         } else {
-            row.style.display = "block";
+            row.style.display = "list-item";
         }
     }
 }
 
-window.addEventListener("load", function () {
+doSetup(function() {
     let hide_accomplished = document.getElementById(
         "hide_accomplished_goals"
     );
@@ -282,7 +294,7 @@ function style_goal_links(style) {
 
 
 /* Toggle for link decoration */
-window.addEventListener("load", function () {
+doSetup(function() {
     let decoration_toggle = document.getElementById(
         "decorate_goal_links"
     );
@@ -413,7 +425,7 @@ function reorganize_goals(grouping_policy) {
 }
 
 // Reorganize goals on load + on selector change if there is one
-window.addEventListener("load", function () {
+doSetup(function() {
     let grouping_policy_selector = document.getElementById(
         "grouping_policy"
     );
@@ -711,7 +723,7 @@ function sticky_bin_ancestor(element) {
 var TOTAL_FLOATABLE_BINS;
 
 // Set up for sticky bins:
-window.addEventListener("load", function () {
+doSetup(function() {
     let bin_index = 0;
     for (let context of document.querySelectorAll(".sticky_context")) {
         // Observe each sticky bin context
@@ -776,7 +788,7 @@ window.addEventListener("load", function () {
  *-----------------*/
 
 /* Decorate links to rubric items and snippets */
-window.addEventListener("load", function () {
+doSetup(function() {
     // Find all links that link to rubric items or snippets and add some
     // decorations to them based on what they link to. For
     // evaluated goals, we add a goal status marker, while for
