@@ -14,12 +14,12 @@ import wavesynth
 m = opt.testFunction(wavesynth.printTrack)
 
 arpeggio.arpeggio(wavesynth.C4, 0.5)
-m.case().expectOutputLines(
-    "a 0.5s keyboard note at C4",
-    "and a 0.5s keyboard note at E4",
-    "and a 0.5s keyboard note at G4",
-    "and a 0.5s keyboard note at E4",
-    "and a 0.5s keyboard note at C4",
+m.case().checkPrintedLines(
+    "a 0.5s keyboard note at C4 (60% vol)",
+    "and a 0.5s keyboard note at E4 (60% vol)",
+    "and a 0.5s keyboard note at G4 (60% vol)",
+    "and a 0.5s keyboard note at E4 (60% vol)",
+    "and a 0.5s keyboard note at C4",  # missing volume value
 )
 
 # Only one test
