@@ -218,14 +218,10 @@ else:
 for bug_nr in [1, 2, 3, 5, 11]:
     meta.expect(
         "failed",
-        "behavior",
-        "core",
         f"bug #{bug_nr}$"
     )
 
 meta.expect(
     "partial",
-    "behavior",
-    "core",
     "bug #4$"
 )
