@@ -277,10 +277,9 @@ def evaluation_key(course, semester, username, pset, phase, task):
     )
 
 
-def exercise_key(course, semester, username, exercise):
+def old_exercise_key(course, semester, username, exercise):
     """
-    The Redis key for the outcomes-list-history for a particular
-    exercise, submitted by a user for a particular course/semester.
+    Old-format exercises key.
     """
     return redis_key(
         ':'.join(
@@ -288,15 +287,16 @@ def exercise_key(course, semester, username, exercise):
         )
     )
 
-
-def best_exercises_key(course, semester, username):
+def exercise_key(course, semester, username, exercise, category):
     """
-    The Redis key for the per-exercise-best-outcomes-list data for a
-    particular user in a course/semester.
+    The Redis key for the outcomes-list-history for a particular
+    exercise, submitted by a user for a particular course/semester.
+    These are further split by category: no-credit, partial-credit, and
+    full-credit exercises are stored in different history lists.
     """
     return redis_key(
         ':'.join(
-            [course, semester, "exercises", username]
+            [course, semester, "outcomes", username, exercise, category]
         )
     )
 
@@ -1358,12 +1358,55 @@ def set_evaluation(
     return True
 
 
-def fetch_outcomes(course, semester, username, exercise):
+def fetch_old_outcomes(course, semester, username, exercise):
+    """
+    Fetches old outcomes for the given course/semester/uesrname/exercise.
+    """
+    # Redis key to use
+    key = old_exercise_key(course, semester, username, exercise)
+
+    try:
+        exists = _REDIS.exists(key)
+    except Exception:
+        flask.flash("Error checking for outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to check for outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    # Return None without making a fuss if the key just doesn't exist
+    if not exists:
+        return None
+
+    try:
+        responseJSON = _REDIS.get(key)
+        info = json.loads(responseJSON)
+    except Exception:
+        flask.flash("Error fetching or decoding outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to fetch outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    return info
+
+def fetch_outcomes(course, semester, username, exercise, category):
     """
     Fetches the outcomes-list information for the given
-    user/course/semester/exercise. The result will be a list of
-    dictionaries each representing a single submission, in chronological
-    order. Each will have the following keys:
+    user/course/semester/exercise/category. The 'category' should be one
+    of the strings 'full', 'partial', or 'none'. The result will be a
+    list of dictionaries each representing a single submission, in
+    chronological order. Each will have the following keys:
 
     - "submitted_at" - A time string (see
         `potluck.time_utils.timestring`) indicating when the list of
@@ -1377,16 +1420,20 @@ def fetch_outcomes(course, semester, username, exercise):
     - "status": A status string indicating the overall exercise status,
         determined by callers to `save_outcomes`.
     - "credit": A number indicating how much credit (0-1) was earned for
-        this outcome. Note that the best-credit outcome is retained in
-        the best-outcomes dictionary, with ties broken towards
-        more-recent submissions.
+        this outcome. May also be `None` in some cases.
+    - "group_credit": A credit number that determines how much credit is
+        earned towards this exercise group. Will be 0 when "credit" is
+        `None`. If the category is 'none', this will always be 0, if the
+        category is 'partial' it will be greater than 0 and less than 1,
+        and if the category is 'full', it will be 1. This number does NOT
+        account for timeliness (yet; see `fetch_best_outcomes`).
 
     Returns None instead of a list if there is no information for that
-    user/exercise yet, or if an error is encountered while trying to
-    access that information.
+    user/exercise/category yet, or if an error is encountered while
+    trying to access that information.
     """
     # Redis key to use
-    key = exercise_key(course, semester, username, exercise)
+    key = exercise_key(course, semester, username, exercise, category)
 
     try:
         exists = _REDIS.exists(key)
@@ -1424,17 +1471,29 @@ def fetch_outcomes(course, semester, username, exercise):
     return info
 
 
-def fetch_best_outcomes(course, semester, username):
+def fetch_best_outcomes(
+    course,
+    semester,
+    username,
+    exercise,
+    deadline,
+    late_fraction
+):
     """
-    Fetches the best outcomes-list information for all exercises for the
-    given user/course/semester. The result will be a dictionary which
-    maps exercise IDs to dictionaries each with the following keys:
+    Fetches the best outcome information for a particular
+    course/semester/user/exercise. To do that, it needs to know what
+    that user's current deadline is, and what credit multiplier to apply
+    to late submissions. The deadline should be given as a
+    `datetime.datetime` object.
+
+    The result will be a dictionary with the following keys:
 
     - "submitted_at" - A time string (see
         `potluck.time_utils.timestring`) indicating when the list of
         outcomes was submitted.
     - "on_time" - A boolean indicating whether the submission came in on
-        time or not.
+        time or not, based on the deadline given and the submission
+        time.
     - "authors" - A list of usernames for participating authors.
     - "outcomes" - A list of 3-tuple outcomes, which contain a boolean
         for success/failure followed by tag and message strings (see
@@ -1445,53 +1504,145 @@ def fetch_best_outcomes(course, semester, username):
         the outcomes which passed/failed.
     - "credit" - A number indicating how much credit this exercise is
         worth (higher = better) or possibly `None` if there is some issue
-        with the submission (like wrong # of outcomes).
+        with the submission (like wrong # of outcomes). This does not take
+        timeliness into account.
+    - "group_credit" - A credit number that accounts for timeliness, and
+        which will always be a number (it's 0 when `credit` would be `None`).
 
-    Only the IDs of exercises the user has submitted at least once will
-    appear in the base dictionary.
-
-    Returns None if an error is encountered, but otherwise returns at
-    least an empty dictionary, including when the specified user hasn't
-    submitted anything yet.
+    Returns None if an error is encountered, or if the user has no
+    submissions for that exercise.
     """
-    # Redis key to use
-    key = best_exercises_key(course, semester, username)
-
-    try:
-        exists = _REDIS.exists(key)
-    except Exception:
-        flask.flash("Error checking for best-outcomes info.")
-        tb = potluck.html_tools.string_traceback()
-        print(
-            "Failed to check for best-outcomes info at '{}':\n{}".format(
-                key,
-                tb
-            ),
-            file=sys.stderr
+    best = None
+    fulls = fetch_outcomes(course, semester, username, exercise, 'full')
+    if fulls is None:
+        fulls = []
+    for submission in reversed(fulls):  # iterate backwards chronologically
+        when = potluck.time_utils.time_from_timestring(
+            submission["submitted_at"]
         )
-        return None
+        if when <= deadline:
+            submission["on_time"] = True
+            submission["group_credit"] = submission.get(
+                "group_credit",
+                submission.get("credit", 0)
+            )
+        else:
+            submission["on_time"] = False
+            submission["group_credit"] = late_fraction * submission.get(
+                "group_credit",
+                submission.get("credit", 0)
+            )
 
-    # Return an empty dictionary without making a fuss if the key just
-    # doesn't exist
-    if not exists:
-        return {}
+        if best is None or submission["group_credit"] >= best["group_credit"]:
+            best = submission
+            # It this one gets full credit; no need to look for better
+            if submission["group_credit"] == 1:
+                break
 
-    try:
-        allJSON = _REDIS.hgetall(key)
-        result = {key: json.loads(allJSON[key]) for key in allJSON}
-    except Exception:
-        flask.flash("Error fetching or decoding best-outcomes info.")
-        tb = potluck.html_tools.string_traceback()
-        print(
-            "Failed to fetch best-outcomes info at '{}':\n{}".format(
-                key,
-                tb
-            ),
-            file=sys.stderr
+    # Only look at partials if we didn't find a full-credit best
+    # submission
+    if best is None or best["group_credit"] < 1:
+        partials = fetch_outcomes(
+            course,
+            semester,
+            username,
+            exercise,
+            'partial'
         )
-        return None
+        if partials is None:
+            partials = []
+        for submission in partials:
+            when = potluck.time_utils.time_from_timestring(
+                submission["submitted_at"]
+            )
+            if when <= deadline:
+                submission["on_time"] = True
+                submission["group_credit"] = submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
+            else:
+                submission["on_time"] = False
+                submission["group_credit"] = late_fraction * submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
 
-    return result
+            if (
+                best is None
+             or submission["group_credit"] >= best["group_credit"]
+            ):
+                best = submission
+
+    # Only look at no-credit submissions if we have no full or
+    # partial-credit submissions at all.
+    if best is None:
+        nones = fetch_outcomes(
+            course,
+            semester,
+            username,
+            exercise,
+            'none'
+        )
+        if nones is not None:
+            # Always take chronologically last one; none of these are worth
+            # any credit anyways.
+            submission = nones[-1]
+            when = potluck.time_utils.time_from_timestring(
+                submission["submitted_at"]
+            )
+            if when <= deadline:
+                submission["on_time"] = True
+                submission["group_credit"] = submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
+            else:
+                submission["on_time"] = False
+                submission["group_credit"] = late_fraction * submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
+
+    # Try legacy info
+    if best is None or best["group_credit"] < 1:
+        legacy = fetch_old_outcomes(
+            course,
+            semester,
+            username,
+            exercise
+        )
+        if legacy is None:
+            legacy = []
+        for submission in legacy:
+            old_on_time = submission.get("on_time", True)
+            # figure out unpenalized credit if it had been marked late
+            if old_on_time is False:
+                submission["group_credit"] /= late_fraction
+
+            when = potluck.time_utils.time_from_timestring(
+                submission["submitted_at"]
+            )
+            if when <= deadline:
+                submission["on_time"] = True
+                submission["group_credit"] = submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
+            else:
+                submission["on_time"] = False
+                submission["group_credit"] = late_fraction * submission.get(
+                    "group_credit",
+                    submission.get("credit", 0)
+                )
+
+            if (
+                best is None
+             or submission["group_credit"] >= best["group_credit"]
+            ):
+                best = submission
+
+    return best  # might still be None
 
 
 def save_outcomes(
@@ -1504,8 +1655,7 @@ def save_outcomes(
     codeBlocks,
     status,
     credit,
-    group_credit,
-    on_time
+    group_credit
 ):
     """
     Saves a list of outcomes for a specific exercise submitted by a
@@ -1518,19 +1668,17 @@ def save_outcomes(
     filename string and a code string (the filename could also elsehow
     identify the source code was derived from). The `status` value should
     be a string describing the status of the submission, while the credit
-    value should be a number where a higher number indicates a better
-    submission (to be stored preferentially as the "best" submission so
-    far if it's better than the old one). The `on_time` value should be a
-    boolean indicating whether the outcomes should count as 'on-time' or
-    not.
+    value should be a number in [0, 1] where a higher number indicates a
+    better submission.
 
     The list of outcomes and associated code blocks is added to the
-    record of all such lists submitted for that exercise by that user, as
-    a dictionary with the following slots:
+    record of all such lists submitted for that exercise by that user,
+    categorized as 'none', 'partial', or 'full' depending on whether the
+    credit value is 0, between 0 and 1, or 1. It will be stored as a
+    dictionary with the following slots:
 
     - "submitted_at": the current time, as a string (see
         `potluck.time_utils.timestring`).
-    - "on_time": The `on_time` boolean provided.
     - "authors": The list of authors.
     - "outcomes": The list of outcomes.
     - "code": The list of filename/code-string pairs.
@@ -1538,17 +1686,23 @@ def save_outcomes(
     - "credit": The credit number.
     - "group_credit": The credit number for counting group credit.
 
-    It will also replace any old outcomes list which is listed in the
-    best-outcomes-per-exercise info for the submitting user if it's
-    better than or the same as (in terms of credit) the best one
-    previously recorded there. If `credit` is set to None, it will count
-    as -1 for this purpose.
-
     Returns True if it succeeds or False if it encounters some sort of
     error.
     """
+    category = 'none'
+    if group_credit > 1:
+        raise ValueError(
+            "Invalid group_credit value '{}' (must be <= 1).".format(
+                group_credit
+            )
+        )
+    elif group_credit == 1:
+        category = 'full'
+    elif group_credit > 0:
+        category = 'partial'
+
     # Redis key to use
-    key = exercise_key(course, semester, username, exercise)
+    key = exercise_key(course, semester, username, exercise, category)
 
     # Generate a timestamp for the info
     timestring = potluck.time_utils.timestring()
@@ -1558,7 +1712,8 @@ def save_outcomes(
         course,
         semester,
         username,
-        exercise
+        exercise,
+        category
     )
     if recorded_outcomes is None:
         recorded_outcomes = []
@@ -1566,7 +1721,6 @@ def save_outcomes(
     # Here's the info we store
     info = {
         "submitted_at": timestring,
-        "on_time": on_time,
         "authors": authors,
         "outcomes": outcomes,
         "code": codeBlocks,
@@ -1595,55 +1749,6 @@ def save_outcomes(
         )
         return False
 
-    # Also store info in best-outcomes map if it is good
-    best_key = best_exercises_key(course, semester, username)
-    # TODO: We should pipeline this with the other write to try to more
-    # completely avoid situations where the first succeeds and the second
-    # fails!
-    try:
-        old_json = _REDIS.hget(best_key, exercise)
-        if old_json is None:
-            old_credit = -1
-        else:
-            old_best = json.loads(old_json)
-            old_credit = old_best.get(
-                "group_credit",
-                # for older data, group_credit isn't present
-                old_best.get("credit", -1)
-            )
-    except Exception:
-        flask.flash("Failed to read best-outcomes info!")
-        tb = potluck.html_tools.string_traceback()
-        print(
-            "Failed to read best-outcomes info at '{}':\n{}".format(
-                best_key,
-                tb
-            ),
-            file=sys.stderr
-        )
-        return False
-
-    # Missing credit value counts as -1 credit.
-    if group_credit is None:
-        group_credit = -1
-
-    # Only record a new best-outcome-for-this-exercise if the current
-    # group credit value is >= the old one.
-    if group_credit >= old_credit:
-        try:
-            _REDIS.hset(best_key, exercise, json.dumps(info))
-        except Exception:
-            flask.flash("Failed to write best-outcomes info!")
-            tb = potluck.html_tools.string_traceback()
-            print(
-                "Failed to write best-outcomes info at '{}':\n{}".format(
-                    best_key,
-                    tb
-                ),
-                file=sys.stderr
-            )
-            return False
-
     return True
 
 
